import numpy as np
from himena import WidgetDataModel, StandardType
from himena.standards.model_meta import ImageMeta, ArrayAxis
from himena.testing import install_plugin
import pytest

@pytest.fixture(scope="session", autouse=True)
def init_pytest(request):
    install_plugin("himena-image")

@pytest.fixture(scope="function")
def image_data():
    rng = np.random.default_rng(0)
    return WidgetDataModel(
        value=rng.normal(size=(4, 5, 2, 6, 5)),
        type=StandardType.IMAGE,
        metadata=ImageMeta(
            axes=[
                ArrayAxis(name="t", scale=0.2, unit="sec"),
                ArrayAxis(name="z", scale=1.5, unit="um"),
                ArrayAxis(name="c", labels=["green", "red"]),
                ArrayAxis(name="y", scale=0.9, unit="um"),
                ArrayAxis(name="x", scale=0.9, unit="um"),
            ],
            channel_axis=2,
            is_rgb=False,
        ),
    )



import pytest
from himena.widgets import MainWindow


@pytest.mark.parametrize(
    "command",
    [
        "himena-image:0-filter-basic:gaussian_filter",
        "himena-image:0-filter-basic:median_filter",
        "himena-image:0-filter-basic:mean_filter",
        "himena-image:1-filter-variance:std_filter",
        "himena-image:1-filter-variance:coef_filter",
        "himena-image:2-filter-comp:dog_filter",
        "himena-image:2-filter-comp:doh_filter",
        "himena-image:2-filter-comp:log_filter",
        "himena-image:2-filter-comp:laplacian_filter",
    ],
)
def test_filter(make_himena_ui, image_data, command: str):
    ui: MainWindow = make_himena_ui(backend="mock")
    win = ui.add_data_model(image_data)
    ui.exec_action(
        command,
        model_context=win.to_model(),
        with_params={},
    )



from __future__ import annotations

from functools import singledispatch
from typing import Callable, TypeVar
import warnings
import numpy as np
from numpy.typing import NDArray
from scipy import ndimage as ndi

from himena import WidgetDataModel, Parametric, StandardType
from himena.consts import MenuId
from himena.widgets import append_result
from himena.plugins import register_function, configure_gui
from himena.standards import roi, model_meta
from himena.data_wrappers import wrap_array, ArrayWrapper


@register_function(
    title="Measure At Current ROI",
    types=StandardType.IMAGE,
    menus=[MenuId.TOOLS_IMAGE_ROI, "/model_menu/roi"],
    keybindings="M",
    command_id="himena-image:roi-measure-current",
)
def roi_measure_current(model: WidgetDataModel):
    arr = wrap_array(model.value)
    if not isinstance(meta := model.metadata, model_meta.ImageMeta):
        raise ValueError("Image must have an ImageMeta.")
    roi = meta.current_roi
    if roi is None:
        raise ValueError("No ROI selected.")
    current_indices = meta.current_indices
    if current_indices is None:
        raise ValueError("`current_indices` is not set.")
    current_slice = tuple(slice(None) if i is None else i for i in current_indices)
    arr_slice = arr.get_slice(current_slice)
    target = slice_array(roi, arr_slice)
    metrics = METRICS_SHARED | METRICS_ADDITIONAL.get(type(roi), {})
    output: dict[str, float] = {}
    for metric_name, func in metrics.items():
        result = func(roi, target)
        if isinstance(result, np.integer):
            output[metric_name] = int(result)
        elif isinstance(result, np.floating):
            output[metric_name] = float(result)
        else:
            output[metric_name] = result
    append_result(output)
    return None


@register_function(
    title="Measure ROIs ...",
    types=StandardType.IMAGE,
    menus=[MenuId.TOOLS_IMAGE_ROI, "/model_menu/roi"],
    run_async=True,
    command_id="himena-image:roi-measure",
)
def roi_measure(model: WidgetDataModel) -> Parametric:
    # TODO: length, area. etc
    metrics_choices = list(METRICS_SHARED.keys())
    arr = wrap_array(model.value)
    if not isinstance(meta := model.metadata, model_meta.ImageMeta):
        raise ValueError("Image must have an ImageMeta.")
    if axes := meta.axes:
        axis_names = [axis.name for axis in axes[:-2]]
    else:
        axis_names = [f"axis_{i}" for i in range(arr.ndim - 2)]
    axes_choices = [(axis_name, i) for i, axis_name in enumerate(axis_names)]

    @configure_gui(
        metrics={"choices": metrics_choices, "widget_type": "Select"},
        along={"choices": axes_choices, "widget_type": "Select"},
    )
    def run_measure(
        metrics: list[str],
        along: list[int],
        pivot: bool = True,
    ) -> WidgetDataModel:
        if isinstance(meta.rois, roi.RoiListModel):
            rois = meta.rois
        else:
            rois = meta.rois()
        if len(rois) == 0:
            raise ValueError("No ROIs to measure.")

        ndindex_shape = tuple(arr.shape[i] for i in along)
        funcs = [METRICS_SHARED[metric] for metric in metrics]

        # check name collision between axis names and metrics names
        for metric in metrics:
            if metric in axis_names:
                warnings.warn(
                    f"Name collision between axis names and metrics: {metric!r}",
                    UserWarning,
                    stacklevel=2,
                )
        out: dict[str, list] = {}
        for along_i in along:
            axis_name = axis_names[along_i]
            out[axis_name] = []
        if pivot:
            # initialize result dict
            for metric in metrics:
                for each_roi in rois:
                    out[f"{metric}_{each_roi.name}"] = []
            for sl in np.ndindex(ndindex_shape):
                for sl_i, axis_name in zip(sl, axis_names):
                    out[axis_name].append(sl_i)
                for indices, each_roi in rois.iter_with_indices():
                    target = _prep_sliced_array(arr, each_roi, sl, along, indices)
                    for func, metric in zip(funcs, metrics):
                        out[f"{metric}_{each_roi.name}"].append(func(each_roi, target))
        else:
            out["name"] = []
            for metric in metrics:
                out[metric] = []
            for sl in np.ndindex(ndindex_shape):
                for indices, each_roi in rois.iter_with_indices():
                    target = _prep_sliced_array(arr, each_roi, sl, along, indices)
                    out["name"].append(each_roi.name)
                    for sl_i, axis_name in zip(sl, axis_names):
                        out[axis_name].append(sl_i)
                    for func, metric in zip(funcs, metrics):
                        out[metric].append(func(each_roi, target))
        return WidgetDataModel(
            value=out,
            type=StandardType.DATAFRAME,
            title=f"Results of {model.title}",
        )

    return run_measure


def _prep_sliced_array(
    arr: ArrayWrapper,
    each_roi: roi.RoiModel,
    sl: tuple[int, ...],
    along: list[int],
    indices: list[int],
):
    sl_placeholder = list(indices)
    for i, along_i in enumerate(along):
        sl_placeholder[along_i] = sl[i]
    arr_slice = arr.get_slice(tuple(sl_placeholder))
    return slice_array(each_roi, arr_slice)


@singledispatch
def slice_array(r: roi.RoiModel, arr_nd: np.ndarray):
    """Transfrom array from (N, ..., Y, X) to (N, ..., S)

    S is the number of pixels in the ROI."""
    mask = r.to_mask(arr_nd.shape)
    return arr_nd[..., mask]


@slice_array.register
def _(r: roi.RectangleRoi, arr_nd: np.ndarray):
    bb = r.bbox().adjust_to_int("inner")
    arr = arr_nd[..., bb.top : bb.bottom, bb.left : bb.right]
    return arr.reshape(*arr.shape[:-2], arr.shape[-2] * arr.shape[-1])


@slice_array.register
def _(r: roi.EllipseRoi, arr_nd: np.ndarray):
    bb = r.bbox().adjust_to_int("inner")
    arr = arr_nd[..., bb.top : bb.bottom, bb.left : bb.right]
    _yy, _xx = np.indices(arr.shape[-2:])
    mask = (_yy - r.y) ** 2 / r.height**2 + (_xx - r.x) ** 2 / r.width**2 <= 1
    return arr[..., mask]


@slice_array.register
def _(r: roi.PointRoi2D, arr_nd: np.ndarray):
    out = ndi.map_coordinates(arr_nd, [[r.y], [r.x]], order=1, mode="nearest")
    return out


@slice_array.register
def _(r: roi.PointsRoi2D, arr_nd: np.ndarray):
    coords = np.stack([r.ys, r.xs], axis=1)
    out = ndi.map_coordinates(arr_nd, coords, order=1, mode="nearest")
    return out


@slice_array.register
def _(r: roi.LineRoi, arr_nd: np.ndarray):
    xs, ys = r.arange()
    return _slice_array_along_line(arr_nd, xs, ys)


@slice_array.register
def _(r: roi.SegmentedLineRoi, arr_nd: np.ndarray):
    xs, ys = r.arange()
    return _slice_array_along_line(arr_nd, xs, ys)


def _slice_array_along_line(arr_nd: NDArray[np.number], xs, ys):
    coords = np.stack([ys, xs], axis=0)
    out = np.empty(arr_nd.shape[:-2] + (coords.shape[1],), dtype=np.float32)
    for sl in np.ndindex(arr_nd.shape[:-2]):
        arr_2d = arr_nd[sl]
        out[sl] = ndi.map_coordinates(arr_2d, coords, order=1, mode="nearest")
    return out


_Roi = TypeVar("_Roi", bound=roi.RoiModel)
# _MetricsType takes a Roi and 1D array slice and returns a float
_MetricsType = Callable[[_Roi, NDArray[np.number], NDArray[np.number]], float]
# Supported additional metrics for each ROI type
METRICS_SHARED: dict[str, _MetricsType[roi.RoiModel]] = {
    "mean": lambda roi, ar_sl: np.mean(ar_sl),
    "std": lambda roi, ar_sl: np.std(ar_sl),
    "min": lambda roi, ar_sl: np.min(ar_sl),
    "max": lambda roi, ar_sl: np.max(ar_sl),
    "sum": lambda roi, ar_sl: np.sum(ar_sl),
    "median": lambda roi, ar_sl: np.median(ar_sl),
    "area": lambda roi, ar_sl: ar_sl.size,
}
_METRICS_LINE: dict[str, _MetricsType[roi.LineRoi]] = {
    "length": lambda roi, ar_sl: roi.length(),
    "angle": lambda roi, ar_sl: roi.angle(),
}
_METRICS_SEGMENTED_LINE: dict[str, _MetricsType[roi.SegmentedLineRoi]] = {
    "length": lambda roi, ar_sl: roi.length(),
}
_METRICS_RECTANGLE: dict[str, _MetricsType[roi.RectangleRoi]] = {
    "area": lambda roi, ar_sl: roi.area(),
    "width": lambda roi, ar_sl: roi.width,
    "height": lambda roi, ar_sl: roi.height,
}
_METRICS_ELLIPSE: dict[str, _MetricsType[roi.EllipseRoi]] = {
    "area": lambda roi, ar_sl: roi.area(),
    "width": lambda roi, ar_sl: roi.width,
    "height": lambda roi, ar_sl: roi.height,
    "circumference": lambda roi, ar_sl: roi.circumference(),
    "eccentricity": lambda roi, ar_sl: roi.eccentricity(),
}
_METRICS_ROTATED_RECTANGLE: dict[str, _MetricsType[roi.RotatedRectangleRoi]] = {
    "area": lambda roi, ar_sl: roi.area(),
    "width": lambda roi, ar_sl: roi.width,
    "lenght": lambda roi, ar_sl: roi.length(),
    "angle": lambda roi, ar_sl: roi.angle(),
}

METRICS_ADDITIONAL: dict[type[roi.RoiModel], dict[str, _MetricsType[roi.LineRoi]]] = {
    roi.LineRoi: _METRICS_LINE,
    roi.SegmentedLineRoi: _METRICS_SEGMENTED_LINE,
    roi.RectangleRoi: _METRICS_RECTANGLE,
    roi.EllipseRoi: _METRICS_ELLIPSE,
    roi.RotatedRectangleRoi: _METRICS_ROTATED_RECTANGLE,
}



from himena import WidgetDataModel, Parametric, StandardType
from himena.plugins import register_function, configure_gui, configure_submenu
from himena_image.utils import (
    make_dims_annotation,
    image_to_model,
    model_to_image,
    norm_dims,
)

MENUS = ["image/analyze/fft", "/model_menu/analyze/fft"]

configure_submenu(MENUS, title="Fourier transform")


@register_function(
    title="FFT ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    run_async=True,
    command_id="himena-image:fft",
)
def fft(model: WidgetDataModel) -> Parametric:
    """Fast Fourier transformation of an image."""

    @configure_gui(dimension={"choices": make_dims_annotation(model)}, preview=True)
    def run_fft(
        origin_in_center: bool = True,
        double_precision: bool = False,
        dimension=2,
        is_previewing: bool = False,
    ) -> WidgetDataModel:
        img = model_to_image(model, is_previewing)
        out = img.fft(
            shift=origin_in_center,
            double_precision=double_precision,
            dims=norm_dims(dimension, img.axes),
        )
        return image_to_model(out, orig=model, is_previewing=is_previewing)

    return run_fft


@register_function(
    title="IFFT ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    run_async=True,
    command_id="himena-image:ifft",
)
def ifft(model: WidgetDataModel) -> Parametric:
    """Inverse fast Fourier transformation of a frequency image."""

    @configure_gui(
        dimension={"choices": make_dims_annotation(model)},
        preview=True,
    )
    def run_ifft(
        return_real: bool = True,
        origin_in_center: bool = True,
        double_precision: bool = False,
        dimension=2,
        is_previewing: bool = False,
    ) -> WidgetDataModel:
        img = model_to_image(model, is_previewing)
        out = img.ifft(
            real=return_real,
            shift=origin_in_center,
            double_precision=double_precision,
            dims=norm_dims(dimension, img.axes),
        )
        return image_to_model(out, orig=model, is_previewing=is_previewing)

    return run_ifft


@register_function(
    title="Power spectrum ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    run_async=True,
    command_id="himena-image:power_spectrum",
)
def power_spectrum(model: WidgetDataModel) -> Parametric:
    """Compute the power spectrum of an image."""

    @configure_gui(
        dimension={"choices": make_dims_annotation(model)},
        norm={"label": "normalize maximum to 1"},
        zero_norm={"label": "normalize zero frequency to 0"},
        preview=True,
    )
    def run_power_spectrum(
        origin_in_center: bool = True,
        norm: bool = False,
        zero_norm: bool = False,
        double_precision: bool = False,
        dimension=2,
        is_previewing: bool = False,
    ) -> WidgetDataModel:
        img = model_to_image(model, is_previewing)
        out = img.power_spectra(
            shift=origin_in_center,
            double_precision=double_precision,
            norm=norm,
            zero_norm=zero_norm,
            dims=norm_dims(dimension, img.axes),
        )
        return image_to_model(out, orig=model, is_previewing=is_previewing)

    return run_power_spectrum


@register_function(
    title="Low-pass Filter ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    run_async=True,
    command_id="himena-image:fft-filter:lowpass_filter",
)
def lowpass_filter(model: WidgetDataModel) -> Parametric:
    """Apply a low-pass filter to an image."""

    @configure_gui(
        dimension={"choices": make_dims_annotation(model)},
        cutoff={"min": 0.0, "max": 1.0},
        order={"min": 1},
        preview=True,
    )
    def run_lowpass_filter(
        cutoff: float = 0.2,
        order: int = 2,
        dimension=2,
        is_previewing: bool = False,
    ) -> WidgetDataModel:
        img = model_to_image(model)
        out = img.lowpass_filter(
            cutoff=cutoff, order=order, dims=norm_dims(dimension, img.axes)
        )
        return image_to_model(out, orig=model, is_previewing=is_previewing)

    return run_lowpass_filter


@register_function(
    title="High-pass Filter ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    run_async=True,
    command_id="himena-image:fft-filter:highpass_filter",
)
def highpass_filter(model: WidgetDataModel) -> Parametric:
    """Apply a high-pass filter to an image."""

    @configure_gui(
        dimension={"choices": make_dims_annotation(model)},
        cutoff={"min": 0.0, "max": 1.0},
        order={"min": 1},
        preview=True,
    )
    def run_highpass_filter(
        cutoff: float = 0.2,
        order: int = 2,
        dimension=2,
        is_previewing: bool = False,
    ) -> WidgetDataModel:
        img = model_to_image(model)
        out = img.highpass_filter(
            cutoff=cutoff,
            order=order,
            dims=norm_dims(dimension, img.axes),
        )
        return image_to_model(out, orig=model, is_previewing=is_previewing)

    return run_highpass_filter


@register_function(
    title="Band-pass Filter ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    run_async=True,
    command_id="himena-image:fft-filter:bandpass_filter",
)
def bandpass_filter(model: WidgetDataModel) -> Parametric:
    """Apply a band-pass filter to an image."""

    @configure_gui(
        dimension={"choices": make_dims_annotation(model)},
        cutoff={"min": 0.0, "max": 1.0},
        cuton={"min": 0.0, "max": 1.0},
        preview=True,
    )
    def run_bandpass_filter(
        cuton: float = 0.2,
        cutoff: float = 0.5,
        order: int = 2,
        dimension=2,
        is_previewing: bool = False,
    ) -> WidgetDataModel:
        img = model_to_image(model)
        out = img.bandpass_filter(
            cuton=cuton,
            cutoff=cutoff,
            order=order,
            dims=norm_dims(dimension, img.axes),
        )
        return image_to_model(out, orig=model, is_previewing=is_previewing)

    return run_bandpass_filter



from himena_image.processing import (
    calculate,
    exposure,
    features,
    filters,
    fft,
    transform,
    morphology,
    measure,
    restore,
)

del calculate, exposure, features, filters, fft, transform, morphology, measure, restore



from cmap import Colormap
from himena import WidgetDataModel, Parametric, create_model
from himena.widgets import TabArea
from himena.consts import StandardType
from himena.plugins import register_function, configure_gui
from himena.standards import roi, plotting as hplt
from himena.standards.model_meta import ImageMeta
import numpy as np
from himena.widgets import SubWindow
from himena_image.utils import model_to_image

MENUS = ["image/exposure", "/model_menu/exposure"]


@register_function(
    title="Histogram",
    menus=MENUS,
    types=[StandardType.IMAGE],
    command_id="himena-image:exposure:histogram",
    run_async=True,
)
def histogram(model: WidgetDataModel) -> Parametric:
    """Show histogram of the image."""
    from skimage.exposure import histogram as skimage_histogram

    img = model_to_image(model)
    meta = _cast_meta(model)
    if meta.is_rgb:
        caxis = -1
        colors = ["red", "green", "blue"]
    elif meta.channel_axis is None:
        caxis = None
        colors = ["gray"]
    else:
        caxis = meta.channel_axis
        colors = [Colormap(chn.colormap or "gray")(0.5).hex for chn in meta.channels]

    @configure_gui(
        bins={"min": 2, "max": 2048},
        normalize={"label": "Normalize the histogram by the sum"},
        channel_axis={"bind": caxis},
        colors={"bind": colors},
    )
    def run_histogram(
        bins: int = 256,
        normalize: bool = True,
        channel_axis: int | None = None,
        colors: list[str] = (),
    ) -> WidgetDataModel:
        """Run histogram."""
        for ith, color in enumerate(colors):
            if channel_axis is not None:
                img_slice = img.value.take(ith, axis=channel_axis)
            else:
                img_slice = img.value
            hist, bin_center = skimage_histogram(
                img_slice, nbins=bins, normalize=normalize
            )
            fig = hplt.figure()
            fig.plot(bin_center, hist, color=color)
        fig.axes.x.label = "Intensity"
        fig.axes.y.label = "Frequency"
        return create_model(
            fig,
            type=StandardType.PLOT,
            title=f"Histogram of {model.title}",
        )

    return run_histogram


@register_function(
    title="Auto contrast at selection",
    menus=MENUS,
    types=[StandardType.IMAGE],
    command_id="himena-image:exposure:auto-contrast-selection",
)
def auto_contrast_at_selection(win: SubWindow):
    """Auto contrast the image."""

    model = win.to_model()
    img = model_to_image(model)
    meta = _cast_meta(model)
    sl = tuple(slice(None) if i is None else i for i in meta.current_indices)
    img_slice = img.value[sl]
    if meta.is_rgb:
        img_slice = np.mean(img_slice, axis=-1)
    if isinstance(cur_roi := meta.current_roi, roi.Roi2D):
        mask = cur_roi.to_mask(img_slice.shape)
        hist = img_slice[mask].ravel()
        min_val = hist.min()
        max_val = hist.max()
    else:
        raise ValueError("Current ROI is not a 2D ROI.")
    if meta.channel_axis is None:
        i_channel = 0
    else:
        i_channel = meta.channel_axis
    meta.channels[i_channel].contrast_limits = (min_val, max_val)
    win.update_model(
        model.with_metadata(meta.model_copy(update={"channels": meta.channels}))
    )


@register_function(
    title="Propagate contrast ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    command_id="himena-image:exposure:propagate-contrast",
)
def propagate_contrast(model: WidgetDataModel, tab: TabArea):
    """Propagate contrast to other images in the same tab area."""

    meta = _cast_meta(model)
    clim_map = {
        channel.colormap: channel.contrast_limits
        for channel in meta.channels
        if channel.colormap is not None
    }
    for win in tab:
        if win.model_type() != StandardType.IMAGE:
            continue
        model = win.to_model()
        this_meta = _cast_meta(model)
        if this_meta.is_rgb:
            continue
        new_channels = []
        for chn in this_meta.channels:
            if new_clim := clim_map.get(chn.colormap, None):
                new_channels.append(
                    chn.model_copy(update={"contrast_limits": new_clim})
                )
            else:
                new_channels.append(chn)
        new_meta = this_meta.model_copy(update={"channels": new_channels})
        model = model.with_metadata(new_meta)
        win.update_model(model)
    return None


def _cast_meta(model: WidgetDataModel) -> ImageMeta:
    """Cast the metadata to ImageMeta."""
    if not isinstance(model.metadata, ImageMeta):
        raise TypeError("Metadata is not of type ImageMeta.")
    return model.metadata



from ndv import DataWrapper
from typing import Any, Hashable, Mapping, Sequence
import numpy as np
from himena.types import WidgetDataModel
from himena.standards.model_meta import ImageMeta
from enum import Enum, auto


class ModelDataWrapper(DataWrapper):
    def __init__(self, model: WidgetDataModel):
        super().__init__(model.value)
        if not isinstance(meta := model.metadata, ImageMeta):
            raise ValueError("Invalid metadata")
        self._meta = meta
        self._type = model.type
        self._complex_conversion = ComplexConversionRule.ABS

    @classmethod
    def supports(cls, obj: Any) -> bool:
        return isinstance(obj, WidgetDataModel)

    @property
    def dims(self) -> tuple[str, ...]:
        if axes := self._meta.axes:
            return tuple(a.name for a in axes)
        return tuple(f"axis_{i}" for i in range(len(self._data.shape)))

    @property
    def coords(self) -> Mapping[Hashable, Sequence]:
        """Return the coordinates for the data."""
        return {d: range(s) for d, s in zip(self.dims, self.data.shape)}

    def isel(self, indexers: Mapping[int, int | slice]) -> np.ndarray:
        """Select a slice from a data store using (possibly) named indices."""
        import dask.array as da

        sl = [slice(None)] * len(self._data.shape)
        for k, v in indexers.items():
            sl[k] = v
        out = self._data[tuple(sl)]
        if isinstance(out, da.Array):
            out = out.compute()
        assert isinstance(out, np.ndarray)
        if out.dtype.kind == "b":
            return out.astype(np.uint8)
        elif out.dtype.kind == "c":
            return self._complex_conversion.apply(out)
        return out

    def sizes(self):
        if axes := self._meta.axes:
            names = [a.name for a in axes]
        else:
            names = list(range(len(self._data.shape)))
        return dict(zip(names, self._data.shape))


class ComplexConversionRule(Enum):
    ABS = auto()
    REAL = auto()
    IMAG = auto()
    PHASE = auto()
    LOG_ABS = auto()

    def apply(self, data: np.ndarray) -> np.ndarray:
        if self == ComplexConversionRule.ABS:
            return np.abs(data)
        elif self == ComplexConversionRule.REAL:
            return data.real
        elif self == ComplexConversionRule.IMAG:
            return data.imag
        elif self == ComplexConversionRule.PHASE:
            return np.angle(data)
        elif self == ComplexConversionRule.LOG_ABS:
            return np.log(np.abs(data) + 1e-10)
        raise ValueError(f"Unknown complex conversion rule: {self}")



from himena import WidgetDataModel, Parametric
from himena.consts import StandardType
from himena.plugins import register_function, configure_gui
from himena_image.consts import PaddingMode, InterpolationOrder
from himena_image.utils import (
    make_dims_annotation,
    image_to_model,
    model_to_image,
    norm_dims,
)

MENUS = ["image/analyze/restore", "/model_menu/analyze/restore"]


@register_function(
    title="Track drift ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    command_id="himena-image:restore:track_drift",
    run_async=True,
)
def track_drift(model: WidgetDataModel) -> Parametric:
    """Track drift in the image."""
    img = model_to_image(model)
    along_default, along_choices = _along_default_and_choices(img.axes)

    @configure_gui(
        along={"choices": along_choices, "value": along_default},
        dimension={"choices": make_dims_annotation(model)},
    )
    def run_track_drift(
        along: str,
        max_shift: float | None = None,
        upsample_factor: int = 10,
    ) -> WidgetDataModel:
        img = model_to_image(model)
        out = img.track_drift(
            along=along,
            upsample_factor=upsample_factor,
            max_shift=max_shift,
        )

        return WidgetDataModel(
            value=out,
            type=StandardType.DATAFRAME,
        )

    return run_track_drift


@register_function(
    title="Drift correction ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    command_id="himena-image:restore:drift_correction",
    run_async=True,
)
def drift_correction(model: WidgetDataModel) -> Parametric:
    """Correct drift in the image."""
    img = model_to_image(model)
    along_default, along_choices = _along_default_and_choices(img.axes)

    @configure_gui(
        along={"choices": along_choices, "value": along_default},
        dimension={"choices": make_dims_annotation(model)},
    )
    def run_drift_correction(
        along: str,
        reference: str = "",
        zero_ave: bool = True,
        max_shift: float | None = None,
        order: InterpolationOrder = 1,
        mode: PaddingMode = "constant",
        cval: float = 0.0,
        dimension: int = 2,
    ) -> WidgetDataModel:
        img = model_to_image(model)
        out = img.drift_correction(
            ref=reference or None,
            zero_ave=zero_ave,
            along=along,
            max_shift=max_shift,
            mode=mode,
            cval=cval,
            order=order,
            dims=norm_dims(dimension, img.axes),
        )
        return image_to_model(out, orig=model)

    return run_drift_correction


def _along_default_and_choices(axes) -> tuple[str, list[str]]:
    along_choices = [str(a) for a in axes]
    if "t" in along_choices:
        along_default = "t"
    elif "z" in along_choices:
        along_default = "z"
    else:
        along_default = along_choices[0]
    return along_default, along_choices


@register_function(
    title="Richardson-Lucy deconvolution ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    command_id="himena-image:restore-deconv:lucy-deconv",
    run_async=True,
)
def lucy(model: WidgetDataModel) -> Parametric:
    """Restore image using poin spread function by Richardson-Lucy's method"""

    @configure_gui(
        psf={"types": [StandardType.IMAGE]},
        dimension={"choices": make_dims_annotation(model)},
    )
    def run_lucy(
        psf: WidgetDataModel,
        niter: int = 50,
        dimension: int = 2,
        eps: float = 1e-5,
    ) -> WidgetDataModel:
        img = model_to_image(model)
        out = img.lucy(psf, niter=niter, eps=eps, dims=norm_dims(dimension, img.axes))
        return image_to_model(out, orig=model)

    return run_lucy


@register_function(
    title="Richardson-Lucy TV deconvolution ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    command_id="himena-image:restore-deconv:lucy-tv-deconv",
    run_async=True,
)
def lucy_tv(model: WidgetDataModel) -> Parametric:
    """Restore image using poin spread function by Richardson-Lucy's method with total
    variance regularization."""

    @configure_gui(
        psf={"types": [StandardType.IMAGE]},
        dimension={"choices": make_dims_annotation(model)},
    )
    def run_lucy_tv(
        psf: WidgetDataModel,
        niter: int = 50,
        dimension: int = 2,
        lmd: float = 1.0,
        tol: float = 1e-3,
        eps: float = 1e-5,
    ) -> WidgetDataModel:
        img = model_to_image(model)
        out = img.lucy_tv(
            psf,
            niter=niter,
            lmd=lmd,
            tol=tol,
            eps=eps,
            dims=norm_dims(dimension, img.axe),
        )
        return image_to_model(out, orig=model)

    return run_lucy_tv



from typing import Annotated, Literal

PaddingMode = Literal["reflect", "constant", "nearest", "mirror", "wrap"]
InterpolationOrder = Annotated[
    int,
    {
        "choices": [("nearest", 0), ("linear", 1), ("cubic", 3)],
        "widget_type": "ComboBox",
    },
]



from typing import Annotated, Literal
import impy as ip

from himena import WidgetDataModel, Parametric
from himena.consts import StandardType
from himena.plugins import register_function, configure_gui
from himena.standards.model_meta import ImageMeta
from himena_image.consts import PaddingMode
from himena_image.utils import (
    make_dims_annotation,
    model_to_image,
    image_to_model,
    norm_dims,
)

MENUS = ["image/analyze/filter", "/model_menu/analyze/filter"]


@register_function(
    title="Gaussian Filter ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    command_id="himena-image:0-filter-basic:gaussian_filter",
    run_async=True,
)
def gaussian_filter(model: WidgetDataModel) -> Parametric:
    """Apply a Gaussian filter to the image."""

    @configure_gui(
        dimension={"choices": make_dims_annotation(model)},
        preview=True,
    )
    def run_gaussian_filter(
        sigma: Annotated[float, {"min": 0.0}] = 1.0,
        dimension: int = 2,
        is_previewing: bool = False,
    ) -> WidgetDataModel:
        img = model_to_image(model, is_previewing)
        out = img.gaussian_filter(sigma=sigma, dims=norm_dims(dimension, img.axes))
        return image_to_model(out, orig=model, is_previewing=is_previewing)

    return run_gaussian_filter


@register_function(
    title="Median Filter ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    command_id="himena-image:0-filter-basic:median_filter",
    run_async=True,
)
def median_filter(model: WidgetDataModel) -> Parametric:
    """Apply a median filter to the image."""

    @configure_gui(
        dimension={"choices": make_dims_annotation(model)},
        preview=True,
    )
    def run_median_filter(
        radius: Annotated[float, {"min": 0.0}] = 1.0,
        mode: PaddingMode = "reflect",
        cval: float = 0,
        dimension: int = 2,
        is_previewing: bool = False,
    ) -> WidgetDataModel:
        img = model_to_image(model, is_previewing)
        out = img.median_filter(
            radius, mode=mode, cval=cval, dims=norm_dims(dimension, img.axes)
        )
        return image_to_model(out, orig=model, is_previewing=is_previewing)

    return run_median_filter


@register_function(
    title="Mean Filter ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    command_id="himena-image:0-filter-basic:mean_filter",
    run_async=True,
)
def mean_filter(model: WidgetDataModel) -> Parametric:
    """Apply a mean filter to the image."""

    @configure_gui(dimension={"choices": make_dims_annotation(model)}, preview=True)
    def run_mean_filter(
        radius: Annotated[float, {"min": 0.0}] = 1.0,
        mode: PaddingMode = "reflect",
        cval: float = 0,
        dimension: int = 2,
        is_previewing: bool = False,
    ) -> WidgetDataModel:
        img = model_to_image(model, is_previewing)
        out = img.mean_filter(
            radius, mode=mode, cval=cval, dims=norm_dims(dimension, img.axes)
        )
        return image_to_model(out, orig=model, is_previewing=is_previewing)

    return run_mean_filter


@register_function(
    title="Min Filter ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    command_id="himena-image:0-filter-basic:min_filter",
    run_async=True,
)
def min_filter(model: WidgetDataModel) -> Parametric:
    """Apply a minimum filter to the image."""

    @configure_gui(dimension={"choices": make_dims_annotation(model)}, preview=True)
    def run_min_filter(
        radius: Annotated[float, {"min": 0.0}] = 1.0,
        mode: PaddingMode = "reflect",
        cval: float = 0,
        dimension: int = 2,
        is_previewing: bool = False,
    ) -> WidgetDataModel:
        img = model_to_image(model, is_previewing)
        out = img.min_filter(
            radius, mode=mode, cval=cval, dims=norm_dims(dimension, img.axes)
        )
        return image_to_model(out, orig=model, is_previewing=is_previewing)

    return run_min_filter


@register_function(
    title="Max Filter ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    command_id="himena-image:0-filter-basic:max_filter",
    run_async=True,
)
def max_filter(model: WidgetDataModel) -> Parametric:
    """Apply a maximum filter to the image."""

    @configure_gui(dimension={"choices": make_dims_annotation(model)}, preview=True)
    def run_max_filter(
        radius: Annotated[float, {"min": 0.0}] = 1.0,
        mode: PaddingMode = "reflect",
        cval: float = 0,
        dimension: int = 2,
        is_previewing: bool = False,
    ) -> WidgetDataModel:
        img = model_to_image(model, is_previewing)
        out = img.max_filter(
            radius, mode=mode, cval=cval, dims=norm_dims(dimension, img.axes)
        )
        return image_to_model(out, orig=model, is_previewing=is_previewing)

    return run_max_filter


@register_function(
    title="STD Filter ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    command_id="himena-image:1-filter-variance:std_filter",
    run_async=True,
)
def std_filter(model: WidgetDataModel) -> Parametric:
    """Standard deviation filter."""

    @configure_gui(dimension={"choices": make_dims_annotation(model)}, preview=True)
    def run_std_filter(
        radius: Annotated[float, {"min": 0.0}] = 1.0,
        mode: PaddingMode = "reflect",
        cval: float = 0,
        dimension: int = 2,
        is_previewing: bool = False,
    ) -> WidgetDataModel:
        img = model_to_image(model, is_previewing)
        out = img.std_filter(
            radius, mode=mode, cval=cval, dims=norm_dims(dimension, img.axes)
        )
        return image_to_model(out, orig=model, is_previewing=is_previewing)

    return run_std_filter


@register_function(
    title="Coef Filter ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    command_id="himena-image:1-filter-variance:coef_filter",
    run_async=True,
)
def coef_filter(model: WidgetDataModel) -> Parametric:
    """Coefficient of variation filter."""

    @configure_gui(dimension={"choices": make_dims_annotation(model)}, preview=True)
    def run_coef_filter(
        radius: Annotated[float, {"min": 0.0}] = 1.0,
        mode: PaddingMode = "reflect",
        cval: float = 0,
        dimension: int = 2,
        is_previewing: bool = False,
    ) -> WidgetDataModel:
        img = model_to_image(model, is_previewing)
        out = img.coef_filter(
            radius, mode=mode, cval=cval, dims=norm_dims(dimension, img.axes)
        )
        return image_to_model(out, orig=model, is_previewing=is_previewing)

    return run_coef_filter


@register_function(
    title="Difference of Gaussian (DoG) Filter ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    command_id="himena-image:2-filter-comp:dog_filter",
    run_async=True,
)
def dog_filter(model: WidgetDataModel) -> Parametric:
    @configure_gui(dimension={"choices": make_dims_annotation(model)}, preview=True)
    def run_dog_filter(
        sigma_low: Annotated[float, {"min": 0.0}] = 1.0,
        sigma_high: Annotated[float, {"min": 0.0}] = 1.6,
        dimension: int = 2,
        is_previewing: bool = False,
    ) -> WidgetDataModel:
        img = model_to_image(model, is_previewing)
        out = img.dog_filter(sigma_low, sigma_high, dims=norm_dims(dimension, img.axes))
        return image_to_model(out, orig=model, is_previewing=is_previewing)

    return run_dog_filter


@register_function(
    title="Laplacian Filter ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    command_id="himena-image:2-filter-comp:laplacian_filter",
    run_async=True,
)
def laplacian_filter(model: WidgetDataModel) -> Parametric:
    @configure_gui(dimension={"choices": make_dims_annotation(model)}, preview=True)
    def run_laplacian_filter(
        radius: Annotated[int, {"min": 1}] = 1,
        dimension: int = 2,
        is_previewing: bool = False,
    ) -> WidgetDataModel:
        img = model_to_image(model, is_previewing)
        out = img.laplacian_filter(radius=radius, dims=norm_dims(dimension, img.axes))
        return image_to_model(out, orig=model, is_previewing=is_previewing)

    return run_laplacian_filter


@register_function(
    title="Difference of Hessian (DoH) Filter ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    command_id="himena-image:2-filter-comp:doh_filter",
    run_async=True,
)
def doh_filter(model: WidgetDataModel) -> Parametric:
    @configure_gui(dimension={"choices": make_dims_annotation(model)}, preview=True)
    def run_doh_filter(
        sigma: Annotated[float, {"min": 0.0}] = 1.0,
        dimension: int = 2,
        is_previewing: bool = False,
    ) -> WidgetDataModel:
        img = model_to_image(model, is_previewing)
        out = img.doh_filter(sigma, dims=norm_dims(dimension, img.axes))
        return image_to_model(out, orig=model, is_previewing=is_previewing)

    return run_doh_filter


@register_function(
    title="Laplacian of Gaussian (LoG) Filter ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    command_id="himena-image:2-filter-comp:log_filter",
    run_async=True,
)
def log_filter(model: WidgetDataModel) -> Parametric:
    """Apply a Laplacian of Gaussian filter to the image."""

    @configure_gui(dimension={"choices": make_dims_annotation(model)}, preview=True)
    def run_log_filter(
        sigma: Annotated[float, {"min": 0.0}] = 1.0,
        dimension: int = 2,
        is_previewing: bool = False,
    ) -> WidgetDataModel:
        img = model_to_image(model, is_previewing)
        out = img.log_filter(sigma, dims=norm_dims(dimension, img.axes))
        return image_to_model(out, orig=model, is_previewing=is_previewing)

    return run_log_filter


@register_function(
    title="Rolling ball ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    command_id="himena-image:9-filter-other:rolling_ball",
    run_async=True,
)
def rolling_ball(model: WidgetDataModel) -> Parametric:
    """Remove or create a background using the rolling-ball algorithm."""

    @configure_gui(dimension={"choices": make_dims_annotation(model)})
    def run_rolling_ball(
        radius: Annotated[float, {"min": 0.0}] = 30.0,
        dimension: int = 2,
        return_background: bool = False,
    ) -> WidgetDataModel:
        img = model_to_image(model)
        out = img.rolling_ball(
            radius, return_bg=return_background, dims=norm_dims(dimension, img.axes)
        )
        return image_to_model(out, orig=model)

    return run_rolling_ball


@register_function(
    title="Entropy filter ...",
    menus=MENUS,
    command_id="himena-image:9-filter-other:entropy-filter",
    run_async=True,
)
def entropy_filter(model: WidgetDataModel) -> Parametric:
    """Run entropy filter on an image."""

    @configure_gui(dimension={"choices": make_dims_annotation(model)})
    def run_entropy(
        radius: Annotated[float, {"min": 0.0}] = 5.0,
        dimension: int = 2,
    ) -> WidgetDataModel:
        img = model_to_image(model)
        out = img.entropy_filter(radius, dims=norm_dims(dimension, img.axes))
        return image_to_model(out, orig=model)

    return run_entropy


@register_function(
    title="Enhance contrast ...",
    menus=MENUS,
    command_id="himena-image:9-filter-other:enhance-contrast",
    run_async=True,
)
def enhance_contrast(model: WidgetDataModel) -> Parametric:
    """Run enhance-contrast filter on an image."""

    @configure_gui(dimension={"choices": make_dims_annotation(model)})
    def run_enhance_contrast(
        radius: Annotated[float, {"min": 0.0}] = 1.0,
        dimension: int = 2,
    ) -> WidgetDataModel:
        img = model_to_image(model)
        out = img.enhance_contrast(radius, dims=norm_dims(dimension, img.axes))
        return image_to_model(out, orig=model)

    return run_enhance_contrast


@register_function(
    title="Threshold ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    command_id="himena-image:9-filter-other:threshold",
    run_async=True,
)
def threshold(model: WidgetDataModel) -> Parametric:
    """Binarize the image using a threshold value."""
    from skimage.filters import threshold_yen

    img = ip.asarray(model.value)
    if img.dtype.kind == "f":
        wdgt = "FloatSlider"
    elif img.dtype.kind in "ui":
        wdgt = "Slider"
    else:
        raise ValueError(f"Unsupported dtype: {img.dtype}")
    if isinstance(meta := model.metadata, ImageMeta):
        if inds := meta.current_indices:
            value = threshold_yen(img.value[inds], nbins=128)
        else:
            value = img.value.mean()

    thresh_options = {
        "min": img.min(),
        "max": img.max(),
        "value": value,
        "widget_type": wdgt,
    }

    @configure_gui(threshold=thresh_options, preview=True)
    def run_threshold(
        threshold,
        dark_background: bool = True,
        is_previewing: bool = False,
    ) -> WidgetDataModel:
        img = model_to_image(model, is_previewing)
        out = img.threshold(threshold)
        if not dark_background:
            out = ~out
        model_out = image_to_model(out, orig=model, is_previewing=is_previewing)
        return model_out

    return run_threshold


@register_function(
    title="Edge Filter ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    command_id="himena-image:0-filter-basic:edge_filter",
    run_async=True,
)
def edge_filter(model: WidgetDataModel) -> Parametric:
    """Filters for detecting edges in the image."""

    @configure_gui(dimension={"choices": make_dims_annotation(model)}, preview=True)
    def run_edge_filter(
        method: Literal["sobel", "prewitt", "scharr", "farid"],
        dimension: int = 2,
        is_previewing: bool = False,
    ) -> WidgetDataModel:
        img = model_to_image(model, is_previewing)
        out = img.edge_filter(method, dims=norm_dims(dimension, img.axes))
        return image_to_model(out, orig=model, is_previewing=is_previewing)

    return run_edge_filter


@register_function(
    title="Smooth mask ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    command_id="himena-image:9-filter-other:smooth_mask",
    run_async=True,
)
def smooth_mask(model: WidgetDataModel) -> Parametric:
    @configure_gui(
        title="Smooth mask",
        dimension={"choices": make_dims_annotation(model)},
        preview=True,
    )
    def run_smooth_mask(
        sigma: Annotated[float, {"min": 0.0}] = 1.0,
        dilate_radius: Annotated[float, {"min": 0.0}] = 1.0,
        dark_background: bool = True,
        dimension: int = 2,
    ) -> WidgetDataModel:
        img = model_to_image(model)
        out = img.smooth_mask(
            sigma=sigma,
            dilate_radius=dilate_radius,
            mask_light=not dark_background,
            dims=norm_dims(dimension, img.axes),
        )
        return image_to_model(out, orig=model)

    return run_smooth_mask


@register_function(
    title="Kalman Filter ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    command_id="himena-image:9-filter-other:kalman_filter",
    run_async=True,
)
def kalman_filter(model: WidgetDataModel) -> Parametric:
    """Denoise time-series image using the Kalman filter."""
    img = model_to_image(model)
    along_choices = [str(a) for a in img.axes]
    if "t" in along_choices:
        along_default = "t"
    elif "z" in along_choices:
        along_default = "z"
    else:
        along_default = along_choices[0]

    @configure_gui(
        title="Kalman Filter",
        along={"choices": along_choices, "value": along_default},
        dimension={"choices": make_dims_annotation(model)},
        preview=True,
    )
    def run_kalman_filter(
        gain: Annotated[float, {"min": 0.0}] = 0.1,
        noise_var: Annotated[float, {"min": 0.0}] = 0.1,
        along: str = along_default,
        dimension: int = 2,
    ) -> WidgetDataModel:
        out = model_to_image(model).kalman_filter(
            gain=gain,
            noise_var=noise_var,
            dims=norm_dims(dimension, img.axes),
            along=along,
        )
        return image_to_model(out, orig=model)

    return run_kalman_filter



import impy as ip

from himena import StandardType, WidgetDataModel, Parametric
from himena.plugins import register_function, configure_gui
from himena_image.utils import (
    label_to_model,
    make_dims_annotation,
    model_to_image,
    norm_dims,
)

MENUS = ["image/analyze/features", "/model_menu/analyze/features"]


@register_function(
    title="Label ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    run_async=True,
    command_id="himena-image:label",
)
def label(model: WidgetDataModel) -> Parametric:
    @configure_gui(
        connectivity={"choices": [1, 2, 3]},
        dimension={"choices": make_dims_annotation(model)},
    )
    def run_label(
        connectivity: int = 1,
        dimension: int = 2,
    ) -> WidgetDataModel[ip.Label]:
        img = model_to_image(model)
        out = img.label(connectivity=connectivity, dims=norm_dims(dimension, img.axes))
        return label_to_model(out, orig=model)

    return run_label


@register_function(
    title="Peak local maxima ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    run_async=True,
    command_id="himena-image:peak_local_max",
)
def peak_local_max(model: WidgetDataModel) -> Parametric:
    @configure_gui(
        labels={"types": [StandardType.IMAGE_LABELS]},
        dimension={"choices": make_dims_annotation(model)},
    )
    def run_peak_local_max(
        min_distance: float = 1.0,
        percentile: float | None = None,
        topn: int | None = None,
        exclude_border: bool = True,
        labels: WidgetDataModel | None = None,
        topn_per_label: int | None = None,
        dimension: int = 2,
    ) -> WidgetDataModel:
        img = model_to_image(model)
        if labels is not None:
            img.labels = labels
        out = img.peak_local_max(
            min_distance=min_distance,
            percentile=percentile,
            topn=topn or float("inf"),
            topn_per_label=topn_per_label or float("inf"),
            exclude_border=exclude_border,
            use_labels=labels is not None,
            dims=norm_dims(dimension, img.axes),
        )
        df = {k: out[k].to_numpy() for k in out.columns}
        return WidgetDataModel(
            value=df, type=StandardType.DATAFRAME, title=f"Peaks of {model.title}"
        )

    return run_peak_local_max


REGIONPROPS_CHOICES = [
    "area", "area_bbox", "area_convex", "area_filled", "axis_major_length",
    "axis_minor_length", "bbox", "centroid", "centroid_local", "centroid_weighted",
    "centroid_weighted_local", "coords_scaled", "coords", "eccentricity",
    "equivalent_diameter_area", "euler_number", "extent", "feret_diameter_max",
    "image_intensity", "inertia_tensor", "inertia_tensor_eigvals", "intensity_max",
    "intensity_mean", "intensity_min", "intensity_std", "label", "moments",
    "moments_central", "moments_hu", "moments_normalized", "moments_weighted",
    "moments_weighted_central", "moments_weighted_hu", "moments_weighted_normalized",
    "num_pixels", "orientation", "perimeter", "perimeter_crofton", "solidity",
]  # fmt: skip


@register_function(
    title="Region Properties ...",
    menus=MENUS,
    types=[StandardType.IMAGE],
    run_async=True,
    command_id="himena-image:region_properties",
)
def region_properties(model: WidgetDataModel) -> Parametric:
    """Measure region properties of an image."""

    @configure_gui(
        labels={"types": [StandardType.IMAGE_LABELS]},
        properties={"choices": REGIONPROPS_CHOICES, "widget_type": "Select"},
    )
    def run_region_properties(
        labels: WidgetDataModel[ip.Label],
        properties: list[str] = ["intensity_mean"],
    ) -> WidgetDataModel:
        img = model_to_image(model)
        img.labels = labels.value
        table = img.regionprops(properties=properties)
        dict_ = {}
        for key, prop in table.items():
            dict_[key] = prop.ravel()
        return WidgetDataModel(
            value=dict_,
            type=StandardType.DATAFRAME,
            title=f"Properties of {model.title}",
        )

    return run_region_properties



from __future__ import annotations

from typing import TYPE_CHECKING
from cmap import Colormap
from qtpy import QtWidgets as QtW, QtCore

from ndv import ArrayViewer
from superqt import QEnumComboBox
from himena.types import WidgetDataModel
from himena.standards.model_meta import ImageMeta, ArrayAxis
from himena.plugins import validate_protocol
from himena_image.widgets._wrapper import ComplexConversionRule

if TYPE_CHECKING:
    from ndv.views._qt._array_view import _QArrayViewer


class NDImageViewer(ArrayViewer):
    def __init__(self):
        super().__init__()
        self._control_widget = QtW.QWidget()
        layout = QtW.QHBoxLayout(self._control_widget)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight)
        spacer = QtW.QWidget()
        spacer.setSizePolicy(
            QtW.QSizePolicy.Policy.Expanding, QtW.QSizePolicy.Policy.Expanding
        )
        layout.addWidget(spacer)
        container = QtW.QWidget()
        self._complex_conversion_rule_cbox = QEnumComboBox(
            enum_class=ComplexConversionRule
        )
        self._complex_conversion_rule_cbox.currentEnumChanged.connect(
            self._on_complex_conversion_rule_changed
        )
        layout.addWidget(self._complex_conversion_rule_cbox)
        layout.addWidget(container)

    @validate_protocol
    def update_model(self, model: WidgetDataModel):
        self.data = model
        is_complex = model.value.dtype.kind == "c"
        self._complex_conversion_rule_cbox.setVisible(is_complex)
        meta = model.metadata
        if isinstance(meta, ImageMeta):
            if meta.is_rgb:
                raise ValueError("RGB images are not supported yet")
            for ch, lut in zip(meta.channels, self.display_model.luts.values()):
                lut.cmap = ch.colormap or "gray"
                lut.clims = ch.contrast_limits
        if is_complex:
            self._complex_conversion_rule_cbox.setCurrentEnum(ComplexConversionRule.ABS)

    @validate_protocol
    def to_model(self) -> WidgetDataModel:
        indices = [
            None if isinstance(v, slice) else v
            for v in self.display_model.current_index.values()
        ]
        self.display_model.luts
        return WidgetDataModel(
            value=self.data,
            type=self.model_type(),
            metadata=ImageMeta(
                current_indices=indices,
                axes=[ArrayAxis(name=a) for a in self.data_wrapper.dims],
            ),
        )

    @validate_protocol
    def size_hint(self) -> tuple[int, int]:
        return (320, 400)

    @validate_protocol
    def model_type(self) -> str:
        return self.data_wrapper._type

    @validate_protocol
    def native_widget(self) -> _QArrayViewer:
        return self.widget()

    @validate_protocol
    def control_widget(self):
        return self._control_widget

    def _on_complex_conversion_rule_changed(self, enum_: ComplexConversionRule):
        self.data_wrapper._complex_conversion = enum_
        if enum_ is ComplexConversionRule.PHASE:
            cmap_name = "cmocean:phase"
        else:
            cmap_name = "inferno"
        # for ctrl in self._lut_ctrls.values():
        #     ctrl._cmap.setCurrentColormap(cmap.Colormap(cmap_name))
        # self.refresh()
        for val in self.display_model.luts.values():
            val.cmap = Colormap(cmap_name)
